const debug = require('debug')('meshblu-azure-proxy:server')

const AzureEventHub   = require('azure-event-hubs').Client
const AzureIotHub     = require('azure-iothub').Client
const AzureMessage    = require('azure-iot-common').Message

const MeshbluFirehose = require('meshblu-firehose-socket.io')
const MeshbluHttp     = require('meshblu-http')

const async           = require('async')
const _               = require('lodash')

class Server {
  constructor({ azureIotConnectionString, meshbluConfig }) {
    _.bindAll(this, Object.getOwnPropertyNames(Server.prototype))
    this.ready = false

    this.azureEventHub = AzureEventHub.fromConnectionString(azureIotConnectionString)
    this.azureIotHub = AzureIotHub.fromConnectionString(azureIotConnectionString)
    this.meshbluFirehose = new MeshbluFirehose({meshbluConfig})
    this.meshbluHttp = new MeshbluHttp(meshbluConfig)
    this.uuid = meshbluConfig.uuid

    this.meshbluFirehose.on('message', this.onMeshbluFirehoseMessage)
  }

  run(callback=_.noop) {
    const jobs = [
      this.fetchDeviceMappings,
      this.connectAzureEventHub,
      this.connectAzureIotHub,
      this.connectMeshbluFirehose,
    ]
    async.parallel(jobs, (error) => {
      if (error) return callback(error)
      this.ready = true
      console.log('proxy ready')
      return callback()
    })
  }

  panic(error) {
    console.error(error.stack)
    process.exit(1)
  }

  fetchDeviceMappings(callback) {
    this.meshbluHttp.whoami((error, state) => {
      if (error) return callback(error)
      this.mapDevices(state)
      return callback()
    })
  }

  mapDevices(state) {
    this.meshbluToAzureMap = {}
    this.azureToMeshbluMap = {}
    _.forEach(state.devicePairs, (pair) => {
      this.meshbluToAzureMap[pair.meshblu] = pair.azure
      this.azureToMeshbluMap[pair.azure] = pair.meshblu
    })
    debug('fetched mappings', state.devicePairs)
    debug('meshbluToAzureMap', this.meshbluToAzureMap)
    debug('azureToMeshbluMap', this.azureToMeshbluMap)
  }

  connectAzureEventHub(callback) {
    const client = this.azureEventHub
    const self = this
    client.open()
    .then(client.getPartitionIds.bind(client))
    .then(function (partitionIds) {
      return partitionIds.map((partitionId) => {
        return client.createReceiver('$Default', partitionId, { 'startAfterTime' : Date.now()}).then(function(receiver) {
          receiver.on('errorReceived', self.panic)
          receiver.on('message', self.onEventHubMessage.bind(self))
        })
      })
    })
    .then(() => {
      debug('connected AzureEventHub')
      return callback()
    })
    .catch(self.panic)
  }

  onEventHubMessage(eventHubMessage) {
    if (!this.ready) {
      debug('called onEventHubMessage but not ready')
      return
    }

    const to = eventHubMessage.systemProperties.to
    const toDevice = to.match(/\/devices\/([^\/]*)\/messages\/events/)[1]
    const proxyDevice = this.azureToMeshbluMap[toDevice]

    if (!proxyDevice) {
      console.error(`mapping from azure device ${toDevice} to meshblu does not exist`)
      return
    }

    var body = _.cloneDeep(eventHubMessage.body)
    debug('onEventHubMessage', {body})

    var type = 'message'
    if (_.get(body,'metadata.type')) {
      type = body.metadata.type
    }

    if (type === 'message') {
      return this.sendMeshbluMessage(body.data, proxyDevice)
    }

    if (type === 'update') {
      return this.updateMeshbluDevice(body.data, proxyDevice)
    }

    console.error(`message from azure device ${toDevice} to meshblu device ${proxyDevice} has invalid type ${type}`)
  }

  sendMeshbluMessage(message, as) {
    if (!message.devices) message.devices = [ '*' ]
    this.meshbluHttp.message(message, {as}, (error) => {
      if (error) {
        console.error('meshblu http error:', error.stack)
        return
      }
      debug('meshblu message sent:', as, message)
    })
  }

  updateMeshbluDevice(config, device) {
    this.meshbluHttp.update(device, config, (error) => {
      if (error) {
        console.error('meshblu http error:', error.stack)
        return
      }
      debug('meshblu config updated:', device, config)
    })
  }

  connectAzureIotHub(callback) {
    this.azureIotHub.open(() => {
      debug('connected AzureIotHub')
      return callback()
    })
  }

  connectMeshbluFirehose(callback) {
    this.meshbluFirehose.once('connect', () => {
      debug('connected MeshbluFirehose')
      return callback()
    })
    this.meshbluFirehose.connect()
  }

  onMeshbluFirehoseMessage(message) {
    debug('meshblu message received:', JSON.stringify(message,null,2))
    if (!this.ready) {
      debug('called onMeshbluFirehoseMessage but not ready')
      return
    }

    const toDevice = message.metadata.route[0].to
    const type = message.metadata.route[0].type

    if (type === 'configure.sent' && toDevice === this.uuid) {
      return this.mapDevices(message.data)
    }

    const proxyDevice = this.meshbluToAzureMap[toDevice]
    if (!proxyDevice) {
      console.error(`mapping from meshblu device ${toDevice} to azure does not exist`)
      return
    }

    const azureMessage = new AzureMessage(message)
    this.azureIotHub.send(proxyDevice, azureMessage)
  }
}

module.exports = Server
